// file      : xsd/cxx-cli/elements.hxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#ifndef CXX_CLI_ELEMENTS_HXX
#define CXX_CLI_ELEMENTS_HXX

#include <cult/types.hxx>
#include <cult/containers/deque.hxx>
#include <cult/containers/vector.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

#include <locale>
#include <ostream>

namespace CXX_CLI
{
  using std::endl;

  using namespace Cult;

  namespace SemanticGraph = XSDFrontend::SemanticGraph;
  namespace Traversal = XSDFrontend::Traversal;

  typedef WideString String;


  // Exceptions.
  //

  struct NoNamespaceMapping
  {
    NoNamespaceMapping (SemanticGraph::Path const& file,
                        UnsignedLong line,
                        UnsignedLong column,
                        String const& ns)
        : file_ (file),
          line_ (line),
          column_ (column),
          ns_ (ns)
    {
    }


    SemanticGraph::Path const&
    file () const
    {
      return file_;
    }

    UnsignedLong
    line () const
    {
      return line_;
    }

    UnsignedLong
    column () const
    {
      return column_;
    }

    String const&
    ns () const
    {
      return ns_;
    }

  private:
    SemanticGraph::Path file_;
    UnsignedLong line_;
    UnsignedLong column_;
    String ns_;
  };


  //
  //
  class Context
  {
  public:
    typedef Cult::Containers::Vector<String> NamespaceMapping;

  public:
    Context (std::wostream& o,
             NarrowString const& char_type__,
             Boolean include_with_brackets__,
             NarrowString const& include_prefix__,
             NarrowString const& esymbol__,
             Containers::Vector<NarrowString> const& nsm,
             Boolean trace_namespace_regex_,
             Boolean inline_)
        : os (o),
          scope (scope_),
          inner_name (inner_name_),
          char_type (char_type_),
          L (L_),
          string_type (string_type_),
          include_with_brackets (include_with_brackets_),
          include_prefix (include_prefix_),
          esymbol (esymbol_),
          inl (inl_),
          locale (locale_),
          char_type_ (char_type__),
          L_ (char_type == L"wchar_t" ? L"L" : L""),
          string_type_ (L"System::String"),
          include_with_brackets_ (include_with_brackets__),
          include_prefix_ (include_prefix__),
          esymbol_ (esymbol__),
          inl_ (inline_ ? L"inline\n" : L""),
          locale_ ("C"),
          scope_stack (scope_stack_),
          trace_namespace_regex (trace_namespace_regex_),
          ns_mapping (ns_mapping_)
    {
      // Default mapping.
      //
      ns_mapping_.push_back (
        L"#^.* (.*?/)??"L"(([a-zA-Z_]\\w*)(/[a-zA-Z_]\\w*)*)$#$2#");
      ns_mapping_.push_back (
        L"#^XMLSchema.xsd http://www.w3.org/2001/XMLSchema$#xml_schema#");

      // Custom mappings.
      //
      for (Containers::Vector<NarrowString>::ConstIterator
             i (nsm.begin ()), e (nsm.end ()); i != e; ++i)
      {
        ns_mapping_.push_back (*i);
      }
    }

  protected:
    Context (Context& c)
        : os (c.os),
          scope (c.scope),
          inner_name (c.inner_name),
          char_type (c.char_type),
          L (c.L),
          string_type (c.string_type),
          include_with_brackets (c.include_with_brackets),
          include_prefix (c.include_prefix),
          esymbol (c.esymbol),
          inl (c.inl),
          locale (c.locale),
          scope_stack (c.scope_stack),
          trace_namespace_regex (c.trace_namespace_regex),
          ns_mapping (c.ns_mapping)
    {
    }

    Context (Context& c, std::wostream& o)
        : os (o),
          scope (c.scope),
          inner_name (c.inner_name),
          char_type (c.char_type),
          L (c.L),
          string_type (c.string_type),
          include_with_brackets (c.include_with_brackets),
          include_prefix (c.include_prefix),
          esymbol (c.esymbol),
          inl (c.inl),
          locale (c.locale),
          scope_stack (c.scope_stack),
          trace_namespace_regex (c.trace_namespace_regex),
          ns_mapping (c.ns_mapping)
    {
    }


  public:
    Void
    enter_scope (String const& name)
    {
      scope_stack.push_back (Scope (name));
      update_scope ();
    }

    Void
    enter_class_scope (String const& name)
    {
      scope_stack.push_back (Scope (name, true));
      update_scope ();
    }

    Void
    enter_class_scope (String const& name, SemanticGraph::Scope& s)
    {
      scope_stack.push_back (Scope (name, s, true));
      update_scope ();
    }

    Void
    leave_scope ()
    {
      scope_stack.pop_back ();
      update_scope ();
    }

  private:
    Void
    update_scope ()
    {
      scope.clear ();

      Boolean first (true);

      for (ScopeStack::Iterator i (scope_stack.begin ());
           i != scope_stack.end ();
           ++i)
      {
        // We only qualify names until the namespace level.
        //
        if (first)
          first = false;
        else
          scope += L"::";

        scope += i->name;
      }

      inner_name.clear ();

      if (!scope_stack.empty ())
        inner_name = scope_stack.back ().name;
    }

  public:

    // Escape C++ keywords and illegal characters.
    //
    String
    escape (String const& name) const;

    // Escape C++ keywords and illegal characters. Perform name
    // conflict resolutions.
    //
    String
    id (String const& name, Char const* suffix = "") const
    {
      return id ("", name, suffix);
    }

    String
    id (Char const* prefix, String const& name, Char const* suffix = "") const;

    // Translate XML namespace name to a C++ identifier.
    //
    String
    ns_name (SemanticGraph::Namespace& ns);

    //
    //
    SemanticGraph::Namespace&
    namespace_ (SemanticGraph::Nameable& n);

    // Original schema namespace name.
    //
    String
    xs_ns_name (SemanticGraph::Nameable& ns);


    // Fully-qualified C++ name.
    //
    String
    fq_name (SemanticGraph::Nameable& n);

  public:
    std::wostream& os;
    String& scope;
    String& inner_name;

    String& char_type;
    String& L;                  // string literal prefix
    String& string_type;

    Boolean& include_with_brackets;
    String& include_prefix;

    String& esymbol;
    String& inl;

    std::locale& locale;

  private:
    String scope_;
    String inner_name_;

    String char_type_;
    String L_;
    String string_type_;

    Boolean include_with_brackets_;
    String include_prefix_;

    String esymbol_;
    String inl_;

    std::locale locale_; // I wonder if it's expensive to keep copies?

    struct Scope
    {
      Scope (String const& n, Boolean c = false)
          : name (n), class_scope (c), scope (0)
      {
      }

      Scope (String const& n, SemanticGraph::Scope& s, Boolean c = false)
          : name (n), class_scope (c), scope (&s)
      {
      }

      String name;
      Boolean class_scope;
      SemanticGraph::Scope* scope;
    };

    typedef
    Cult::Containers::Deque<Scope>
    ScopeStack;

    ScopeStack& scope_stack;
    ScopeStack scope_stack_;

  private:
    Boolean trace_namespace_regex;
    NamespaceMapping ns_mapping_;
    NamespaceMapping const& ns_mapping;
  };


  // Usual namespace mapping.
  //
  struct Namespace : Traversal::Namespace
  {
    Namespace (Context& c, Boolean set_scope = false)
        : ctx_ (c), set_scope_ (set_scope)
    {
    }

    virtual Void
    pre (Type& ns);

    virtual Void
    post (Type&);

  private:
    Context& ctx_;
    Boolean set_scope_;
  };


  //
  //
  template <typename X>
  struct Has : X
  {
    Has (Boolean& result)
        : result_ (result)
    {
    }

    virtual Void
    traverse (typename X::Type&)
    {
      result_ = true;
    }

  private:
    Boolean& result_;
  };

  // Checks if scope 'Y' names any of 'X'
  //
  template <typename X, typename Y>
  Boolean
  has (Y& y)
  {
    Boolean result (false);

    Has<X> t (result);

    Traversal::Scope s;
    Traversal::Names n;
    s.edge_traverser (n);
    n.node_traverser (t);

    s.traverse (y);

    return result;
  }


  // Fundamental type mapping helper.
  //
  struct Fundamental : Traversal::Fundamental::Type,
                       Traversal::Fundamental::String,
                       Traversal::Fundamental::NormalizedString,
                       Traversal::Fundamental::Token,
                       Traversal::Fundamental::Name,
                       Traversal::Fundamental::NameToken,
                       Traversal::Fundamental::NCName,
                       Traversal::Fundamental::Id,
                       Traversal::Fundamental::IdRef
  {
    virtual Void
    fundamental_type (SemanticGraph::Fundamental::Type& t) = 0;

    virtual Void
    fundamental_template (SemanticGraph::Fundamental::Type& t) = 0;

    virtual Void
    traverse (SemanticGraph::Fundamental::Type& t)
    {
      fundamental_type (t);
    }

    virtual Void
    traverse (SemanticGraph::Fundamental::String& t)
    {
      fundamental_template (t);
    }

    virtual Void
    traverse (SemanticGraph::Fundamental::NormalizedString& t)
    {
      fundamental_template (t);
    }

    virtual Void
    traverse (SemanticGraph::Fundamental::Token& t)
    {
      fundamental_template (t);
    }

    virtual Void
    traverse (SemanticGraph::Fundamental::Name& t)
    {
      fundamental_template (t);
    }

    virtual Void
    traverse (SemanticGraph::Fundamental::NameToken& t)
    {
      fundamental_template (t);
    }

    virtual Void
    traverse (SemanticGraph::Fundamental::NCName& t)
    {
      fundamental_template (t);
    }

    virtual Void
    traverse (SemanticGraph::Fundamental::Id& t)
    {
      fundamental_template (t);
    }

    virtual Void
    traverse (SemanticGraph::Fundamental::IdRef& t)
    {
      fundamental_template (t);
    }
  };


  // Anonymous feedback via belongs edge.
  //
  struct AnonymousBase : Traversal::Element,
                         Traversal::Attribute
  {
    AnonymousBase (Traversal::NodeDispatcherBase& d1)
        : complex_ (&d1, 0)
    {
      edge_traverser (belongs_);
      belongs_.node_traverser (complex_);
    }

    AnonymousBase (Traversal::NodeDispatcherBase& d1,
                   Traversal::NodeDispatcherBase& d2)
        : complex_ (&d1, &d2)
    {
      edge_traverser (belongs_);
      belongs_.node_traverser (complex_);
    }

    // Hooks.
    //
  public:
    virtual void
    member_pre (SemanticGraph::Member& m)
    {
    }

    virtual void
    member_post (SemanticGraph::Member& m)
    {
    }

    /*
      virtual void
      type_pre (SemanticGraph::Type& t)
      {
      }

      virtual void
      type_post (SemanticGraph::Type& t)
      {
      }
    */

  public:

    virtual Void
    traverse (SemanticGraph::Element& e)
    {
      SemanticGraph::Type& t (e.type ());

      if (!t.named () && !t.context ().count ("seen"))
      {
        t.context ().set ("seen", true);

        member_pre (e);

        Element::belongs (e, belongs_);

        member_post (e);

        t.context ().remove ("seen");
      }
    }

    virtual Void
    traverse (SemanticGraph::Attribute& a)
    {
      SemanticGraph::Type& t (a.type ());

      if (!t.named () && !t.context ().count ("seen"))
      {
        t.context ().set ("seen", true);

        member_pre (a);

        Attribute::belongs (a, belongs_);

        member_post (a);

        t.context ().remove ("seen");
      }
    }

  private:
    struct Complex : Traversal::Complex
    {
      Complex (Traversal::NodeDispatcherBase* d1,
               Traversal::NodeDispatcherBase* d2)
          : d1_ (d1), d2_ (d2)
      {
      }

      virtual Void
      traverse (SemanticGraph::Complex& c)
      {
        //c.context ().set ("name", scope);

        if (d1_)
          d1_->dispatch (c);

        if (d2_)
          d2_->dispatch (c);

        //c.context ().remove ("name");

        //Traversal::Complex::traverse (c);
      }

    private:
      Traversal::NodeDispatcherBase* d1_;
      Traversal::NodeDispatcherBase* d2_;

    } complex_;

    Traversal::Belongs belongs_;
  };
}

#endif  // CXX_CLI_TREE_ELEMENTS_HXX
