// file      : xsd/cxx/tree/tree-inline.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/tree-inline.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

#include <backend-elements/regex.hxx>

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      struct List : Traversal::List, protected virtual Context
      {
        List (Context& c, String const& name_ = L"")
            : Context (c), name (name_)
        {
        }

        virtual Void
        traverse (Type& l)
        {
          if (l.named ())
            name = l.name ();

          String escaped_name (id (name));

          enter_class_scope (escaped_name);
          String scoped_name (scope);
          leave_scope ();

          os << "// " << scoped_name << endl
             << "//" << endl
             << endl;

          // c-tor ()
          //
          os << inl
             << scoped_name << "::" << endl
             << escaped_name << " ()"
             << "{"
             << "}";

          // copy c-tor ()
          //
          os << inl
             << scoped_name << "::" << endl
             << escaped_name << " (class " << escaped_name << " const& o," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c)" << endl
             << ": ::xsd::cxx::tree::simple_type (o, f, c)," << endl
             << "::xsd::cxx::tree::sequence< " <<
            type_name (l.argumented ().type ()) << " > (o, f, c)"
             << "{"
             << "}";
        }

      private:
        String
        type_name (SemanticGraph::Type& t)
        {
          // This type is always named.
          //
          std::wostringstream o;

          MemberTypeName type (*this, o);
          type.dispatch (t);

          return o.str ();
        }

      private:
        String name;
      };


      struct Union : Traversal::Union, protected virtual Context
      {
        Union (Context& c, String const& name_ = L"")
            : Context (c), name (name_)
        {
        }

        virtual Void
        traverse (Type& u)
        {
          if (u.named ())
            name = u.name ();

          String base (L"::xsd::cxx::tree::string< " + char_type + L" >");

          String escaped_name (id (name));

          enter_class_scope (escaped_name);
          String scoped_name (scope);
          leave_scope ();

          os << "// " << scoped_name << endl
             << "//" << endl
             << endl;

          // c-tor (string const&)
          //
          os << inl
             << scoped_name << "::" << endl
             << escaped_name << " (" << string_type << " const& s)" << endl
             << ": " << base << " (s)"
             << "{"
             << "}";

          // copy c-tor ()
          //
          os << inl
             << scoped_name << "::" << endl
             << escaped_name << " (class " << escaped_name << " const& o," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c)" << endl
             << ": " << base << " (o, f, c)"
             << "{"
             << "}";
        }

      private:
        String name;
      };


      struct Enumeration : Traversal::Enumeration,
                           protected virtual Context
      {
        Enumeration (Context& c, String const& name_ = L"")
            : Context (c), name (name_)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          if (e.named ())
            name = e.name ();

          String escaped_name (id (name));
          String escaped_xsd_name (id ("_xsd_", name));
          String escaped_xsd_name_ (id ("_xsd_", name, "_"));

          enter_class_scope (escaped_name, e);
          String scoped_name (scope);
          leave_scope ();

          os << "// " << scoped_name << endl
             << "// " << endl
             << endl;

          // c-tor (_xsd_name)
          //
          os << inl
             << scoped_name << "::" <<  endl
             << escaped_name << " (" << escaped_xsd_name << " v)" << endl
             << ": " << escaped_xsd_name_ << " (v)" << endl
             << "{"
             << "}";


          // copy c-tor
          //
          os << inl
             << scoped_name << "::" <<  endl
             << escaped_name << " (class " << escaped_name << " const& v," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c)" << endl
             << ": ::xsd::cxx::tree::simple_type (v, f, c)," << endl
             << escaped_xsd_name_ << " (v." << escaped_xsd_name_ << ")"
             << "{"
             << "}";

          // operator= (_xsd_name)
          //
          os << inl
             << "class " << scoped_name << "& " << scoped_name << "::" << endl
             << "operator= (" << escaped_xsd_name << " v)"
             << "{"
             << escaped_xsd_name_ << " = v;"
             << "return *this;"
             << "}";

          os << endl;
        }

      private:
        String name;
      };

      struct MemberFunction: Traversal::Member, protected virtual Context
      {
        MemberFunction (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          String name (m.name ());

          if (m.max () != 1)
          {
            // sequence
            //
            String container (iscope_name (name, "container"));
            String q_container (scope + L"::" + container);

            // container const&
            // name () const;
            //
            os << inl
               << q_container << " const& " << scope << "::" << endl
               << id (name) << " () const"
               << "{"
               << "return this->" << id ("_xsd_", name, "_") << " ();"
               << "}";

            // container&
            // name ();
            //
            os << inl
               << q_container << "& " << scope << "::" << endl
               << id (name) << " ()"
               << "{"
               << "return this->" << id ("_xsd_", name, "_") << " ();"
               << "}";

            // void
            // name (container const&);
            //
            os << inl
               << "void " << scope << "::" << endl
               << id (name) << " (" << container << " const& " << id (name) << ")"
               << "{"
               << "this->" << id ("_xsd_", name, "_") <<
              " (" << id (name) << ");"
               << "}";
          }
          else if (m.min () == 0 &&
                   (!m.is_a<SemanticGraph::Attribute> () || !m.default_ ()))
          {
            // optional
            //
            String type (iscope_name (name, "type"));
            String container (iscope_name (name, "container"));
            String q_container (scope + L"::" + container);

            // container const&
            // name () const;
            //
            os << inl
               << q_container << " const& " << scope << "::" << endl
               << id (name) << " () const"
               << "{"
               << "return this->" << id ("_xsd_", name, "_") << " ();"
               << "}";

            // container&
            // name ();
            //
            os << inl
               << q_container << "& " << scope << "::" << endl
               << id (name) << " ()"
               << "{"
               << "return this->" << id ("_xsd_", name, "_") << " ();"
               << "}";

            // void
            // name (type const&);
            //
            os << inl
               << "void " << scope << "::" << endl
               << id (name) << " (" << type << " const& " << id (name) << ")"
               << "{"
               << "this->" << id ("_xsd_", name, "_") <<
              " (" << id (name) << ");"
               << "}";

            // void
            // name (container const&);
            //
            os << inl
               << "void " << scope << "::" << endl
               << id (name) << " (" << container << " const& " << id (name) << ")"
               << "{"
               << "this->" << id ("_xsd_", name, "_") <<
              " (" << id (name) << ");"
               << "}";

            // void
            // name (auto_ptr<type>);
            //
            os << inl
               << "void " << scope << "::" << endl
               << id (name) << " (::std::auto_ptr< " << type << " > " <<
              id (name) << ")"
               << "{"
               << "this->" << id ("_xsd_", name, "_") <<
              " (" << id (name) << ");"
               << "}";
          }
          else
          {
            // one
            //
            String type (iscope_name (name, "type"));
            String q_type (scope + L"::" + type);

            // type const&
            // name () const;
            //
            os << inl
               << q_type << " const& " << scope << "::" << endl
               << id (name) << " () const"
               << "{"
               << "return this->" << id ("_xsd_", name, "_") << " ();"
               << "}";

            // type&
            // name ();
            //
            os << inl
               << q_type << "& " << scope << "::" << endl
               << id (name) << " ()"
               << "{"
               << "return this->" << id ("_xsd_", name, "_") << " ();"
               << "}";

            // void
            // name (type const&);
            //
            os << inl
               << "void " << scope << "::" << endl
               << id (name) << " (" << type << " const& " << id (name) << ")"
               << "{"
               << "this->" << id ("_xsd_", name, "_") <<
              " (" << id (name) << ");"
               << "}";

            // void
            // name (auto_ptr<type>);
            //
            os << inl
               << "void " << scope << "::" << endl
               << id (name) << " (::std::auto_ptr< " << type << " > " <<
              id (name) << ")"
               << "{"
               << "this->" << id ("_xsd_", name, "_") <<
              " (" << id (name) << ");"
               << "}";
          }
        }
      };

      template <typename Complex>
      struct Member : Traversal::Member,
                      protected virtual Context
      {
        Member (Context& c)
            : Context (c),
              member_function_ (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          String name (m.name ());
          SemanticGraph::Type& t (m.type ());

          os << "// " << scope << "::" << id (name) << endl
             << "// " << endl
             << endl;

          // Go after anonymous type first.
          //
          if (!t.named () && !t.context ().count ("seen"))
          {
            Traversal::Belongs belongs_;

            List list (*this, id (name));
            Union union_ (*this, id (name));
            Complex complex (*this, id (name));
            Enumeration enumeration (*this, id (name));

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            String escaped_xsd_name (id ("_xsd_", name));
            String escaped_xsd_name_ (id ("_xsd_", name, "_"));

            enter_scope (escaped_xsd_name);
            enter_scope (escaped_xsd_name_);

            t.context ().set ("seen", true);

            belongs (m, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }

          member_function_.traverse (m);
        }

      private:
        MemberFunction member_function_;
      };


      struct Complex : Traversal::Complex, protected virtual Context
      {
        Complex (Context& c, String const& name_ = L"")
            : Context (c),
              name (name_),
              member_ (c)
        {
          names_ >> member_;
        }

        virtual Void
        traverse (Type& c)
        {
          if (c.named ())
            name = c.name ();

          String escaped_name (id (name));

          enter_class_scope (escaped_name, c);
          String scoped_name (scope);
          leave_scope ();

          os << "// " << scoped_name << endl
             << "// " << endl
             << endl;


          // Go after anonymous types and generate accessors and modifiers
          // first.
          //
          {
            enter_class_scope (escaped_name, c);

            names (c, names_);

            leave_scope ();
          }

          os << endl;
        }

      private:
        String name;

        Traversal::Names names_;
        Member<Complex> member_;
      };


      // Goes after anonymous types via global element. Note that we don't
      // need to go after attributes because they can only be referenced
      // and in that case everything will be generated at the point of
      // reference.
      //
      struct GlobalMember : Traversal::Element,
                            protected virtual Context
      {
        GlobalMember (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          SemanticGraph::Type& t (m.type ());

          if (!t.named ())
          {
            String name (m.name ());

            Traversal::Belongs belongs_;

            List list (*this, id (name));
            Union union_ (*this, id (name));
            Complex complex (*this, id (name));
            Enumeration enumeration (*this, id (name));

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            String escaped_name (id (name));
            String escaped_xsd_name_ (id ("_xsd_", name, "_"));

            enter_scope (escaped_name);
            enter_scope (escaped_xsd_name_);

            t.context ().set ("seen", true);

            belongs (m, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }
        }
      };

      struct Includes : Traversal::Imports,
                        Traversal::Includes,
                        protected virtual Context
      {
        Includes (Context& c, String const& expr)
            : Context (c), expr_ (expr)
        {
        }

        virtual Void
        traverse (SemanticGraph::Imports& i)
        {
          traverse (i.file ());
        }

        virtual Void
        traverse (SemanticGraph::Includes& i)
        {
          traverse (i.file ());
        }

        virtual Void
        traverse (SemanticGraph::Path const& f)
        {
          using BackendElements::Regex::perl_s;

          // Note the use of the portable representation of the path.
          //
          String name (perl_s (String (f.string ()), expr_));

          Char op (include_with_brackets ? '<' : '"');
          Char cl (include_with_brackets ? '>' : '"');

          os << "#include " << op << include_prefix << name << cl << endl
             << endl;
        }

      private:
        String expr_;
      };
    }

    Void
    generate_tree_inline (Context& ctx,
                          SemanticGraph::Schema& tu,
                          String const& expr)
    {
      // Generate includes.
      //
      if (ctx.options.value<CLI::generate_inline> ())
      {
        Traversal::Schema schema;
        Includes includes (ctx, expr);

        schema >> includes;

        schema.dispatch (tu);
      }


      Traversal::Schema schema;
      Traversal::Sources sources;
      Traversal::Names names_ns, names;
      Namespace ns (ctx);

      List list (ctx);
      Union union_ (ctx);
      Complex complex (ctx);
      Enumeration enumeration (ctx);
      GlobalMember member (ctx);

      schema >> sources >> schema;
      schema >> names_ns >> ns >> names;

      names >> list;
      names >> union_;
      names >> complex;
      names >> enumeration;
      names >> member;

      schema.dispatch (tu);
    }
  }
}
