// file      : xsd/cxx/tree/tree-source.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/tree-source.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      struct List : Traversal::List, protected virtual Context
      {
        List (Context& c, String const& name_ = L"")
            : Context (c),
              name (name_)
        {
        }

        virtual Void
        traverse (Type& l)
        {
          if (l.named ())
            name = l.name ();

          String escaped_name (id (name));

          enter_class_scope (escaped_name);
          String scoped_name (scope);
          leave_scope ();

          os << "// " << scoped_name << endl
             << "//" << endl
             << endl;

          // c-tor (element const&)
          //
          os << scoped_name << "::" << endl
             << escaped_name << " (" << xml_element_type << " const& e," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c)" << endl
             << ": ::xsd::cxx::tree::simple_type (e, f, c)," << endl
             << "::xsd::cxx::tree::sequence< " <<
            item_type_name (l.argumented ().type ()) << " > (e, f, c)"
             << "{"
             << "}";

          // c-tor (attribute const&)
          //
          os << scoped_name << "::" << endl
             << escaped_name << " (" << xml_attribute_type << " const& a," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c)" << endl
             << ": ::xsd::cxx::tree::simple_type (a, f, c)," << endl
             << "::xsd::cxx::tree::sequence< " <<
            item_type_name (l.argumented ().type ()) << " > (a, f, c)"
             << "{"
             << "}";

          // c-tor (string const&, element const*)
          //
          os << scoped_name << "::" << endl
             << escaped_name << " (" << string_type << " const& s," << endl
             << xml_element_type << " const* e," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c)" << endl
             << ": ::xsd::cxx::tree::simple_type (s, e, f, c)," << endl
             << "::xsd::cxx::tree::sequence< " <<
            item_type_name (l.argumented ().type ()) << " > (s, e, f, c)"
             << "{"
             << "}";

          // _clone
          //
          os << type_name (l) << "* " << scoped_name << "::" << endl
             << "_clone (::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c) const"
             << "{"
             << "return new class " << id (name) << " (*this, f, c);"
             << "}";

          // Register with type factory map.
          //
          if (l.named ())
          {
            os << "static" << endl
               << "::xsd::cxx::tree::type_factory_initializer< 0, " <<
              char_type << ", " << type_name (l) << " >" << endl
               << id ("_xsd_", name, "_type_factory_init") << " (" << endl
               << L << "\"" << name << " " << xs_ns_name (l) << "\");"
               << endl
               << endl;
          }
        }

      private:
        String
        item_type_name (SemanticGraph::Type& t)
        {
          // This type is always named.
          //
          std::wostringstream o;

          MemberTypeName type (*this, o);
          type.dispatch (t);

          return o.str ();
        }

        String
        type_name (Type& l)
        {
          if (l.named ())
          {
            std::wostringstream o;

            MemberTypeName type (*this, o);
            type.dispatch (l);

            return o.str ();
          }
          else
          {
            enter_class_scope (id (name));
            String scoped_name (scope);
            leave_scope ();

            return scoped_name;
          }
        }

      private:
        String name;
      };


      struct Union : Traversal::Union, protected virtual Context
      {
        Union (Context& c, String const& name_ = L"")
            : Context (c), name (name_)
        {
        }

        virtual Void
        traverse (Type& u)
        {
          if (u.named ())
            name = u.name ();

          String base (L"::xsd::cxx::tree::string< " + char_type + L" >");

          String escaped_name (id (name));

          enter_class_scope (escaped_name);
          String scoped_name (scope);
          leave_scope ();

          os << "// " << scoped_name << endl
             << "//" << endl
             << endl;

          // c-tor (element const&)
          //
          os << scoped_name << "::" << endl
             << escaped_name << " (" << xml_element_type << " const& e," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c)" << endl
             << ": " << base << " (e, f, c)"
             << "{"
             << "}";

          // c-tor (attribute const&)
          //
          os << scoped_name << "::" << endl
             << escaped_name << " (" << xml_attribute_type << " const& a," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c)" << endl
             << ": " << base << " (a, f, c)"
             << "{"
             << "}";

          // c-tor (string const&, element const*)
          //
          os << scoped_name << "::" << endl
             << escaped_name << " (" << string_type << " const& s," << endl
             << xml_element_type << " const* e," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c)" << endl
             << ": " << base << " (s, e, f, c)"
             << "{"
             << "}";

          // _clone
          //
          os << type_name (u) << "* " << scoped_name << "::" << endl
             << "_clone (::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c) const"
             << "{"
             << "return new class " << id (name) << " (*this, f, c);"
             << "}";

          // Register with type factory map.
          //
          if (u.named ())
          {
            os << "static" << endl
               << "::xsd::cxx::tree::type_factory_initializer< 0, " <<
              char_type << ", " << type_name (u) << " >" << endl
               << id ("_xsd_", name, "_type_factory_init") << " (" << endl
               << L << "\"" << name << " " << xs_ns_name (u) << "\");"
               << endl
               << endl;
          }
        }

      private:
        String
        type_name (Type& u)
        {
          if (u.named ())
          {
            std::wostringstream o;

            MemberTypeName type (*this, o);
            type.dispatch (u);

            return o.str ();
          }
          else
          {
            enter_class_scope (id (name));
            String scoped_name (scope);
            leave_scope ();

            return scoped_name;
          }
        }

      private:
        String name;
      };


      // Enumeration mapping.
      //

      struct Enumerator : Traversal::Enumerator, protected virtual Context
      {
        Enumerator (Context& c, String instance_name_)
            : Context (c), instance_name (instance_name_)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          String name (e.name ());

          os << "if (s == " << L << "\"" << name << "\")"
             << "{"
             << instance_name << " = " <<
            inner_name << "::" << id (name) << ";"
             << "}"
             << "else ";
        }

      private:
        String instance_name;
      };


      struct Enumeration : Traversal::Enumeration, protected virtual Context
      {
        Enumeration (Context& c, String name_ = L"")
            : Context (c),
              name (name_)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          if (e.named ())
            name = e.name ();

          String escaped_name (id (name));
          String escaped_xsd_name_ (id ("_xsd_", name, "_"));

          enter_class_scope (escaped_name, e);
          String scoped_name (scope);
          leave_scope ();

          os << "// " << scoped_name << endl
             << "//" << endl
             << endl;


          // c-tor (element const&)
          //
          os << scoped_name << "::" << endl
             << escaped_name << " (" << xml_element_type << " const& e," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c)" << endl
             << ": ::xsd::cxx::tree::simple_type (e, f, c)"
             << "{"
             << string_type << " s (e.value ());"
             << endl;

          {
            enter_class_scope (escaped_name, e);

            Traversal::Names names_enumerator;
            Enumerator enumerator (*this, escaped_xsd_name_);

            names_enumerator >> enumerator;

            names (e, names_enumerator);

            leave_scope ();
          }

          //@@ TODO: expected list
          //
          os << "{"  // else block
             << "throw ::xsd::cxx::tree::unexpected_enumerator < " <<
            char_type << " > (s);"
             << "}"
             << "}";


          // c-tor (attribute const&)
          //
          os << scoped_name << "::" << endl
             << escaped_name << " (" << xml_attribute_type << " const& a," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c)" << endl
             << ": ::xsd::cxx::tree::simple_type (a, f, c)"
             << "{"
             << string_type << " s (a.value ());"
             << endl;

          {
            enter_class_scope (escaped_name, e);

            Traversal::Names names_enumerator;
            Enumerator enumerator (*this, escaped_xsd_name_);

            names_enumerator >> enumerator;

            names (e, names_enumerator);

            leave_scope ();
          }

          //@@ TODO: expected list
          //
          os << "{"  // else block
             << "throw ::xsd::cxx::tree::unexpected_enumerator < " <<
            char_type << " > (s);"
             << "}"
             << "}";


          // c-tor (string const&, element const*)
          //
          os << scoped_name << "::" << endl
             << escaped_name << " (" << string_type << " const& s," << endl
             << xml_element_type << " const* e," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c)" << endl
             << ": ::xsd::cxx::tree::simple_type (s, e, f, c)"
             << "{";

          {
            enter_class_scope (escaped_name, e);

            Traversal::Names names_enumerator;
            Enumerator enumerator (*this, escaped_xsd_name_);

            names_enumerator >> enumerator;

            names (e, names_enumerator);

            leave_scope ();
          }

          //@@ TODO: expected list
          //
          os << "{"  // else block
             << "throw ::xsd::cxx::tree::unexpected_enumerator < " <<
            char_type << " > (s);"
             << "}"
             << "}";


          // _clone
          //
          os << type_name (e) << "* " << scoped_name << "::" << endl
             << "_clone (::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c) const"
             << "{"
             << "return new class " << id (name) << " (*this, f, c);"
            //<< "return new " << type_name (e) << " (*this, f, c);"
             << "}";


          // Register with type factory map.
          //
          if (e.named ())
          {
            os << "static" << endl
               << "::xsd::cxx::tree::type_factory_initializer< 0, " <<
              char_type << ", " << type_name (e) << " >" << endl
               << id ("_xsd_", name, "_type_factory_init") << " (" << endl
               << L << "\"" << name << " " << xs_ns_name (e) << "\");"
               << endl
               << endl;
          }
        }

      private:
        String
        type_name (Type& e)
        {
          if (e.named ())
          {
            std::wostringstream o;

            MemberTypeName type (*this, o);
            type.dispatch (e);

            return o.str ();
          }
          else
          {
            enter_class_scope (id (name));
            String scoped_name (scope);
            leave_scope ();

            return scoped_name;
          }
        }

      private:
        String name;
      };


      // For anonymous types.
      //
      //
      template <typename Complex>
      struct Member : Traversal::Member,
                      protected virtual Context
      {
        Member (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          SemanticGraph::Type& t (m.type ());

          if (!t.named () && !t.context ().count ("seen"))
          {
            String name (m.name ());

            os << "// " << scope << "::" << id (name) << endl
               << "// " << endl
               << endl;

            Traversal::Belongs belongs_;

            List list (*this, id (name));
            Union union_ (*this, id (name));
            Complex complex (*this, id (name));
            Enumeration enumeration (*this, id (name));

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            String escaped_xsd_name (id ("_xsd_", name));
            String escaped_xsd_name_ (id ("_xsd_", name, "_"));

            enter_scope (escaped_xsd_name);
            enter_scope (escaped_xsd_name_);

            t.context ().set ("seen", true);

            belongs (m, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }
        }
      };


      struct Element : Traversal::Element, protected virtual Context
      {
        Element (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          String name (e.name ());
          String tr (id ("_xsd_", name) + L"::traits_"); // traits type name
          String type (id ("_xsd_", name) + L"::type_");

          // Check if we need to handle xsi:type and substitution groups.
          // If this element's type is anonymous or mapped to a fundamental
          // C++ type then we don't need to do anything.
          //
          SemanticGraph::Type& t (e.type ());
          Boolean polymorphic (t.named ());

          if (polymorphic)
          {
            Boolean fund (false);
            IsFundamentalType traverser (fund);
            traverser.dispatch (t);
            polymorphic = !fund;
          }

          os << "// " << name << endl
             << "//" << endl
             << "{";

          if (polymorphic)
          {
            // aCC cannot handle an inline call to type_factory_map_instance.
            //
            os << "::xsd::cxx::tree::type_factory_map< " << char_type <<
              " >& tfm (" << endl
               << "::xsd::cxx::tree::type_factory_map_instance< 0, " <<
              char_type << " > ());"
               << endl
               << "::std::auto_ptr< " << type << " > r (" << endl
               << "tfm.create< " << type << " > (" << endl
               << L << "\"" << name << "\"," << endl
               << L << "\"" <<
              (e.qualified () ? e.namespace_ ().name () : L"") << "\"," << endl
               << (e.global () ? "true" : "false") << "," << endl
               << (e.qualified () ? "true" : "false") << "," << endl
               << "e," << endl
               << "f | ::xsd::cxx::tree::flags::not_root," << endl
               << "this));"
               << endl
               << "if (r.get () != 0)"
               << "{";
          }
          else
          {
            if (e.qualified ())
            {
              os << "if (e.name () == " << L << "\"" << name << "\" && " <<
                "e.namespace_ () == " << L << "\"" <<
                e.namespace_ ().name () << "\")"
                 << "{";
            }
            else
            {
              os << "if (e.name () == " << L << "\"" << name << "\" && " <<
                "e.namespace_ ().empty ())"
                 << "{";
            }
          }


          // Checks.
          //
          if (e.max () != 1)
          {
            // sequence
            //
          }
          else if (e.min () == 0)
          {
            // optional
            //
            os << "if (this->" << id (name) << " ())"
               << "{"
               << "throw ::xsd::cxx::tree::unexpected_element< " <<
              char_type << " > (" << endl
               << "e.name (), e.namespace_ ()," << endl
               << L << "\"\", " << L << "\"\");"
               << "}";
          }
          else
          {
            // one
            //
            os << "if (" << id ("_xsd_", name, "_") << "._present ())"
               << "{"
               << "throw ::xsd::cxx::tree::unexpected_element< " <<
              char_type << " > (" << endl
               << "e.name (), e.namespace_ ()," << endl
               << L << "\"\", " << L << "\"\");"
               << "}";
          }


          if (polymorphic)
          {
            if (e.max () != 1)
            {
              // sequence
              //
              os << "this->" << id (name) << " ().push_back (r);";
            }
            else if (e.min () == 0)
            {
              // optional
              //
              os << "this->" << id (name) << " (r);";
            }
            else
            {
              // one
              //
              os << "this->" << id (name) << " (r);";
            }
          }
          else
          {
            if (e.max () != 1)
            {
              // sequence
              //
              os << "this->" << id (name) << " ().push_back (" << endl
                 << tr << "::create (" << endl
                 << "e, f | ::xsd::cxx::tree::flags::not_root, this));";
            }
            else if (e.min () == 0)
            {
              // optional
              //
              os << "this->" << id (name) << " (" << endl
                 << tr << "::create (" << endl
                 << "e, f | ::xsd::cxx::tree::flags::not_root, this));";
            }
            else
            {
              // one
              //
              os << "this->" << id (name) << " (" << endl
                 << tr << "::create (" << endl
                 << "e, f | ::xsd::cxx::tree::flags::not_root, this));";
            }
          }

          os << "continue;"
             << "}" // if block
             << "}";
        }
      };

      struct ElementTest : Traversal::Element, protected virtual Context
      {
        ElementTest (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          String name (e.name ());

          if (e.max () == 1 && e.min () == 1)
          {
            // one
            //
            os << "if (!" << id ("_xsd_", name, "_") << "._present ())"
               << "{"
               << "throw ::xsd::cxx::tree::expected_element< " <<
              char_type << " > (" << endl
               << L << "\"" << name << "\"," << endl
               << L << "\"" <<
              (e.qualified () ? e.namespace_ ().name () : L"") << "\");"
               << "}";
          }
        }
      };

      struct Attribute : Traversal::Attribute, protected virtual Context
      {
        Attribute (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& a)
        {
          String name (a.name ());
          String tr (id ("_xsd_", name) + L"::traits_"); // traits type name


          if (a.qualified ())
          {
            os << "if (a.name () == " << L << "\"" << name << "\" && " <<
              "a.namespace_ () == " << L << "\"" <<
              a.namespace_ ().name () << "\")"
               << "{";
          }
          else
          {
            os << "if (a.name () == " << L << "\"" << name << "\" && " <<
              "a.namespace_ ().empty ())"
               << "{";
          }

          Boolean fund (false);

          {
            IsFundamentalType traverser (fund);
            traverser.dispatch (a.type ());
          }

          if (fund)
          {
            os << "this->" << id (name) << " (" << endl
               << tr << "::create (" << endl
               << "a, f | ::xsd::cxx::tree::flags::not_root, this));";
          }
          else
          {
            String type (id ("_xsd_", name) + L"::type_");

            os << "::std::auto_ptr< " << type << " > r (" << endl
               << tr << "::create (" << endl
               << "a, f | ::xsd::cxx::tree::flags::not_root, this));"
               << endl
               << "this->" << id (name) << " (r);";
          }

          os << "}"
             << "else ";
        }
      };


      struct AttributeTest : Traversal::Attribute, protected virtual Context
      {
        AttributeTest (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& a)
        {
          String name (a.name ());

          if (!a.optional () || a.default_ ())
          {
            os << "if (!" << id ("_xsd_", name, "_") << "._present ())"
               << "{";

            if (a.default_ ())
              os << "this->" << id (name) << " (" << endl
                 << id ("_xsd_", name) << "::traits_::create< " <<
                char_type << " > (" << endl
                 << string_type << " (" << L << "\"" << a.value () <<
                "\")," << endl
                 << "e," << endl
                 << "f | ::xsd::cxx::tree::flags::not_root," << endl
                 << "this));";
            else
              os << "throw ::xsd::cxx::tree::expected_attribute< " <<
                char_type << " > (" << endl
                 << L << "\"" << name << "\"," << endl
                 << L << "\"" <<
                (a.qualified () ? a.namespace_ ().name () : L"") << "\");";

            os << "}";
          }
        }
      };


      // List of all non-optional members
      //
      struct CtorArgs : Traversal::Complex,
                        Traversal::Enumeration,
                        Traversal::Type,
                        Traversal::Member,
                        protected virtual Context
      {
        CtorArgs (Context& c, String const& name)
            : Context (c), name_ (name)
        {
          complex_ >> inherits_;
          complex_ >> names_;

          inherits_ >> *this;
          names_ >> *this;
        }

        virtual Void
        traverse (SemanticGraph::Type& t)
        {
          os << comma () << name_ << "::member_ const& " << name_;
        }

        virtual Void
        traverse (SemanticGraph::Enumeration& e)
        {
          os << comma () << name_ << "::member_ const& " << name_;
        }

        virtual Void
        traverse (SemanticGraph::Member& m)
        {
          // Note that we are not going to include attributes with
          // default values here. Instead we are going to default-
          // initialize them.
          //
          if (m.min () == 1 && m.max () == 1)
          {
            String name (m.name ());

            String type ();

            os << comma () << iscope_name (name, "type") << " const& " <<
              id ("_xsd_", name);
          }
        }

        Void
        traverse (SemanticGraph::Complex& c)
        {
          first_ = true;
          complex_.traverse (c);
        }

      public:
        String
        comma ()
        {
          return first_ ? (first_ = false, L"") : L",\n";
        }

      private:
        Boolean first_;
        String name_;

      private:
        Traversal::Complex complex_;

        Traversal::Inherits inherits_;
        Traversal::Names names_;
      };


      //
      //
      struct CtorBase : Traversal::Complex,
                        Traversal::Enumeration,
                        Traversal::Type,
                        protected virtual Context
      {
        CtorBase (Context& c, String const& name_)
            : Context (c), name (name_), args_ (c, name_)
        {
        }

        virtual Void
        traverse (SemanticGraph::Type& t)
        {
          os << name;
        }

        virtual Void
        traverse (SemanticGraph::Enumeration& e)
        {
          os << name;
        }

        Void
        traverse (SemanticGraph::Complex& c)
        {
          args_.dispatch (c);
        }

      private:
        struct Args : Traversal::Complex,
                      protected Traversal::Enumeration,
                      protected Traversal::Type,
                      protected Traversal::Member,
                      protected virtual Context
        {
          Args (Context& c, String const& name_)
              : Context (c), name (name_)
          {
            complex_.edge_traverser (inherits_);
            complex_.edge_traverser (names_);

            inherits_.node_traverser (*this);
            names_.node_traverser (*this);
          }

          virtual Void
          traverse (SemanticGraph::Type& t)
          {
            os << comma () << name;
          }

          virtual Void
          traverse (SemanticGraph::Enumeration& e)
          {
            os << comma () << name;
          }

          virtual Void
          traverse (SemanticGraph::Member& m)
          {
            // Note that we are not including attributes with default
            // values here.
            if (m.min () == 1 && m.max () == 1)
            {
              // one
              //
              os << comma () << id ("_xsd_", m.name ());
            }
          }

          Void
          traverse (SemanticGraph::Complex& c)
          {
            first_ = true;

            complex_.traverse (c);
          }

        private:
          String
          comma ()
          {
            return first_ ? (first_ = false, L"") : L",\n";
          }

          Boolean first_;

        private:
          String name;

          Traversal::Complex complex_;

          Traversal::Inherits inherits_;
          Traversal::Names names_;
        } args_;

        String name;
      };


      struct CTorMember : Traversal::Member,
                          protected virtual Context
      {
        CTorMember (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          String name (m.name ());
          //String i (id ("_xsd_", name, "_"));         // instance name
          //String t (id ("_xsd_", name) + L"::type_"); // type name

          if (m.max () != 1)
          {
            // sequence
            //
            os << "," << endl
               << id ("_xsd_", name, "_") << " (" <<
              "::xsd::cxx::tree::flags::not_root, this)";
          }
          else if (m.min () == 0 &&
                   (!m.is_a<SemanticGraph::Attribute> () || !m.default_ ()))
          {
            // optional
            //
            os << "," << endl
               << id ("_xsd_", name, "_") << " (" <<
              "::xsd::cxx::tree::flags::not_root, this)";
          }
          else
          {
            // one
            //

            if (m.min () == 0)
            {
              // This is an attribute with default value. We are going
              // to initialize it to its default value.
              //
              String tr (); // traits type name

              os << "," << endl
                 << id ("_xsd_", name, "_") << " (" << endl
                 << id ("_xsd_", name) << "::traits_::create< " <<
                char_type << " > (" << endl
                 << string_type << " (" << L << "\"" << m.value () <<
                "\")," << endl
                 << "::xsd::cxx::tree::flags::not_root," << endl
                 << "this)," << endl
                 << "::xsd::cxx::tree::flags::not_root," << endl
                 << "this)";
            }
            else
            {
              os << "," << endl
                 << id ("_xsd_", name, "_") << " (" <<
                id ("_xsd_", name) << "," << endl
                 << "::xsd::cxx::tree::flags::not_root," << endl
                 << "this)";
            }
          }

        }
      };


      struct CopyMember : Traversal::Member,
                          protected virtual Context

      {
        CopyMember (Context& c, String const& arg_name_)
            : Context (c), arg_name (arg_name_)
        {
        }

        virtual Void
        traverse (SemanticGraph::Member& m)
        {
          String name (m.name ());

          if (m.max () != 1)
          {
            // sequence
            //

          }
          else if (m.min () == 0 &&
                   (!m.is_a<SemanticGraph::Attribute> () || !m.default_ ()))
          {
            // optional
            //
          }
          else
          {
            //
            // one
          }

          os << "," << endl
             << id ("_xsd_", name, "_") << " (" <<
            arg_name << "." << id ("_xsd_", name, "_") << "," << endl
             << "f | ::xsd::cxx::tree::flags::not_root," << endl
             << "this)";
        }

      private:
        String arg_name;
      };


      // C-tor initializer.
      //
      struct ElementCTorMember : Traversal::Member,
                                 protected virtual Context
      {
        ElementCTorMember (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          String name (m.name ());

          if (m.max () != 1)
          {
            // sequence
            //
          }
          else if (m.min () == 0 &&
                   (!m.is_a<SemanticGraph::Attribute> () || !m.default_ ()))
          {
            // optional
            //
          }
          else
          {
            // one
            //
          }

          os << "," << endl
             << id ("_xsd_", name, "_") <<
            " (f | ::xsd::cxx::tree::flags::not_root, this)";
        }
      };


      //
      //
      struct Complex : Traversal::Complex, protected virtual Context
      {
        Complex (Context& c, String const& name_ = L"")
            : Context (c),
              name (name_),
              member_ (c),
              element_ (c),
              element_test_ (c),
              attribute_ (c),
              attribute_test_ (c),
              ctor_member_ (c),
              element_ctor_member_ (c)
        {
          names_member_ >> member_;

          names_element_ >> element_;
          names_element_test_ >> element_test_;

          names_attribute_ >> attribute_;
          names_attribute_test_ >> attribute_test_;

          ctor_member_names_ >> ctor_member_;

          element_ctor_member_names_ >> element_ctor_member_;
        }


        virtual Void
        traverse (Type& c)
        {
          if (c.named ())
            name = c.name ();

          String escaped_name (id (name));
          String escaped_xsd_name (id ("_xsd_", name));
          String base (escaped_xsd_name + L"::base_"); // base type name

          enter_class_scope (escaped_name, c);
          String scoped_name (scope);
          leave_scope ();

          os << "// " << scoped_name << endl
             << "//" << endl
             << endl;

          // Go after anonymous types first.
          //
          {
            enter_class_scope (escaped_name, c);

            names (c, names_member_);

            leave_scope ();
          }


          // c-tor (all-non-optional-members)
          //
          os << scoped_name << "::" << endl
             << escaped_name << " (";

          {
            enter_class_scope (escaped_name, c);

            CtorArgs ctor_args (*this, escaped_xsd_name);
            ctor_args.dispatch (c);

            leave_scope ();
          }


          os << ")" << endl
             << ": " << base << " (";

          {
            enter_class_scope (escaped_name, c);

            Traversal::Inherits inherits_;
            CtorBase base (*this, escaped_xsd_name);

            inherits_ >> base;

            inherits (c, inherits_);

            leave_scope ();
          }

          os << ")";

          {
            enter_class_scope (escaped_name, c);

            names (c, ctor_member_names_);

            leave_scope ();
          }

          os << "{"
             << "}";


          // copy c-tor
          //
          os << scoped_name << "::" << endl
             << escaped_name << " (class " << escaped_name << " const& " <<
            escaped_xsd_name << "," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c)" << endl
             << ": " << base << " (" << escaped_xsd_name << ", f, c)";

          {
            enter_class_scope (escaped_name, c);

            Traversal::Names names_;
            CopyMember copy_member (*this, escaped_xsd_name);

            names_ >> copy_member;

            names<Complex> (c, names_);

            leave_scope ();
          }

          os << "{"
             << "}";


          //
          //
          Boolean he (has<Traversal::Element> (c));
          Boolean ha (has<Traversal::Attribute> (c));

          // c-tor (xml::element)
          //
          os << scoped_name << "::" << endl
             << escaped_name << " (" << xml_element_type << " const& e," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c)" << endl
             << ": " << base << " (e, f, c)";

          {
            enter_class_scope (escaped_name, c);

            names (c, element_ctor_member_names_);

            leave_scope ();
          }


          os << "{"
             << endl;


          if (he || ha)
          {
            os << parser_type << " p (e);"
               << endl;

            if (he)
            {
              os << "while (p.more_elements ())"
                 << "{"
                 << xml_element_type << " e (p.next_element ());"
                 << endl;

              {
                enter_class_scope (escaped_name, c);

                names (c, names_element_);

                leave_scope ();
              }

              //@@ TODO: expected list
              //
                // @@
                // This doesn't play well with inheritance because we
                // don't expect base's elements.
                //
                //<< "throw ::xsd::cxx::tree::unexpected_element < " <<
                //char_type << " > (n, " << L << "\"" << "\");"

              os << "}"; // while loop

              // Make sure all non-optional elements are set.
              //
              {
                enter_class_scope (escaped_name, c);

                names (c, names_element_test_);

                leave_scope ();
              }
            }

            if (ha)
            {
              os << "while (p.more_attributes ())"
                 << "{"
                 << xml_attribute_type << " a (p.next_attribute ());"
                 << endl;

              {
                enter_class_scope (escaped_name, c);

                names (c, names_attribute_);

                leave_scope ();
              }

              os << "{"  // else block
                // @@
                // This doesn't play well with inheritance because we
                // don't expect base's attributes. Also there are other
                // "special" attributes such as xmlns, etc.
                //
                //<< "throw ::xsd::cxx::tree::unexpected_attribute < " <<
                //char_type << " > (n, " << L << "\"" << "\");"
                 << "}";
              os << "}"; // while loop

              // Make sure all non-optional attributes are set.
              //
              {
                enter_class_scope (escaped_name, c);

                names (c, names_attribute_test_);

                leave_scope ();
              }
            }
          }

          os << "}";

          Boolean r (true);

          struct Traverser : Traversal::Complex
          {
            Traverser (Boolean& v)
                : v_ (v)
            {
              edge_traverser (inherits_);
              inherits_.node_traverser (*this);
            }

            virtual Void
            traverse (SemanticGraph::Complex& c)
            {
              if (has<Traversal::Element> (c) ||
                  has<Traversal::Attribute> (c))
                v_ = false;
              else
                inherits (c, *this);
            }

          private:
            Traversal::Inherits inherits_;
            Boolean& v_;
          } t (r);

          t.dispatch (c);

          if (r)
          {
            // c-tor (xml::attribute)
            //
            os << scoped_name << "::" << endl
               << escaped_name << " (" << xml_attribute_type << " const& a," << endl
               << "::xsd::cxx::tree::flags f," << endl
               << "::xsd::cxx::tree::type* c)" << endl
               << ": " << base << " (a, f, c)"
               << "{"
               << "}";

            // c-tor (string const&, element const*)
            //
            os << scoped_name << "::" << endl
               << escaped_name << " (" << string_type << " const& s," << endl
               << xml_element_type << " const* p," << endl
               << "::xsd::cxx::tree::flags f," << endl
               << "::xsd::cxx::tree::type* c)" << endl
               << ": " << base << " (s, p, f, c)"
               << "{"
               << "}";
          }

          // _clone
          //
          os << type_name (c) << "* " << scoped_name << "::" << endl
             << "_clone (::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c) const"
             << "{"
             << "return new class " << id (name) << " (*this, f, c);"
            //<< "return new " << type_name (c) << " (*this, f, c);"
             << "}";

          // Register with type factory map.
          //
          if (c.named ())
          {
            os << "static" << endl
               << "::xsd::cxx::tree::type_factory_initializer< 0, " <<
              char_type << ", " << type_name (c) << " >" << endl
               << id ("_xsd_", name, "_type_factory_init") << " (" << endl
               << L << "\"" << name << " " << xs_ns_name (c) << "\");"
               << endl
               << endl;
          }
        }

      private:
        String
        type_name (Type& c)
        {
          if (c.named ())
          {
            std::wostringstream o;

            MemberTypeName type (*this, o);
            type.dispatch (c);

            return o.str ();
          }
          else
          {
            enter_class_scope (id (name));
            String scoped_name (scope);
            leave_scope ();

            return scoped_name;
          }
        }

      private:
        String name;

        Member<Complex> member_;
        Traversal::Names names_member_;

        Traversal::Names names_element_;
        Traversal::Names names_element_test_;

        Traversal::Names names_attribute_;
        Traversal::Names names_attribute_test_;

        Element element_;
        ElementTest element_test_;

        Attribute attribute_;
        AttributeTest attribute_test_;

        CTorMember ctor_member_;
        Traversal::Names ctor_member_names_;

        ElementCTorMember element_ctor_member_;
        Traversal::Names element_ctor_member_names_;
      };


      // Goes after anonymous types via global element. Note that we don't
      // need to go after attributes because they can only be referenced
      // and in that case everything will be generated at the point of
      // reference.
      //
      struct GlobalMember : Traversal::Element, protected virtual Context
      {
        GlobalMember (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          SemanticGraph::Type& t (e.type ());

          if (!t.named ())
          {
            String name (e.name ());

            Traversal::Belongs belongs_;

            List list (*this, id (name));
            Union union_ (*this, id (name));
            Complex complex (*this, id (name));
            Enumeration enumeration (*this, id (name));

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            String escaped_name (id (name));
            String escaped_xsd_name_ (id ("_xsd_", name, "_"));

            enter_scope (escaped_name);
            enter_scope (escaped_xsd_name_);

            t.context ().set ("seen", true);

            belongs (e, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }

          if (SemanticGraph::Substitutes* s = e.substitutes ())
          {
            Type& r (s->root ());

            String type_name;

            if (t.named ())
            {
              std::wostringstream o;

              MemberTypeName type (*this, o);
              type.dispatch (t);

              type_name = o.str ();
            }
            else
            {
              String name (id (e.name ()));
              type_name = name + L"::" +
                (name == L"type" ? L"type_" : L"type");
            }

            os << "static" << endl
               << "::xsd::cxx::tree::type_factory_initializer< 0, " <<
              char_type << ", " << type_name << " >" << endl
               << id ("_xsd_", e.name (), "_element_factory_init") << " (" << endl
               << L << "\"" << r.name () << " " << r.namespace_ ().name () <<
              "\", " << endl
               << L << "\"" << e.name () << " " << e.namespace_ ().name () <<
              "\");" << endl
               << endl
               << endl;
          }
        }
      };
    }

    Void
    generate_tree_source (Context& ctx, SemanticGraph::Schema& tu)
    {
      ctx.os << "#include <xsd/cxx/xml/dom/parser.hxx>" << endl
             << "#include <xsd/cxx/tree/type-factory-map.hxx>" << endl
             << endl;

      ctx.os << "namespace _xsd"
             << "{"
             << "static" << endl
             << "::xsd::cxx::tree::type_factory_plate< 0, " <<
        ctx.char_type << " >" << endl
             << "type_factory_plate_init;"
             << "}";

      Traversal::Schema schema;
      Traversal::Sources sources;
      Traversal::Names names_ns, names;
      Namespace ns (ctx);

      List list (ctx);
      Union union_ (ctx);
      Complex complex (ctx);
      Enumeration enumeration (ctx);
      GlobalMember member (ctx);

      schema >> sources >> schema;
      schema >> names_ns >> ns >> names;

      names >> list;
      names >> union_;
      names >> complex;
      names >> enumeration;
      names >> member;

      schema.dispatch (tu);
    }
  }
}
